@@ -32,9 +32,26 @@ module Nokogiri
     # * Nokogiri::XML::Node#next
     # * Nokogiri::XML::Node#previous
     #
-    # You may search this node's subtree using Node#xpath and Node#css
+    #
+    # When printing or otherwise emitting a document or a node (and
+    # its subtree), there are a few methods you might want to use:
+    #
+    # * content, text, inner_text, to_str: emit plaintext
+    #  
+    #   These methods will all emit the plaintext version of your
+    #   document, meaning that entities will be replaced (e.g., "&lt;"
+    #   will be replaced with "<"), meaning that any sanitizing will
+    #   likely be un-done in the output.
+    #
+    # * to_s, to_xml, to_html, inner_html: emit well-formed markup
+    #
+    #   These methods will all emit properly-escaped markup, meaning
+    #   that it's suitable for consumption by browsers, parsers, etc.
+    #
+    # You may search this node's subtree using Searchable#xpath and Searchable#css
     class Node
       include Nokogiri::XML::PP::Node
+      include Nokogiri::XML::Searchable
       include Enumerable
 
       # Element node type, see Nokogiri::XML::Node#element?
@@ -90,130 +107,6 @@ module Nokogiri
         document.decorate(self)
       end
 
-      ###
-      # Search this node for +paths+.  +paths+ can be XPath or CSS, and an
-      # optional hash of namespaces may be appended.
-      # See Node#xpath and Node#css.
-      def search *paths
-        # TODO use         paths, handler, ns, binds = extract_params(paths)
-        ns = paths.last.is_a?(Hash) ? paths.pop :
-          (document.root ? document.root.namespaces : {})
-
-        prefix = "#{implied_xpath_context}/"
-
-        xpath(*(paths.map { |path|
-          path = path.to_s
-          path =~ /^(\.\/|\/|\.\.|\.$)/ ? path : CSS.xpath_for(
-            path,
-            :prefix => prefix,
-            :ns     => ns
-          )
-        }.flatten.uniq) + [ns])
-      end
-      alias :/ :search
-
-      ###
-      # call-seq: xpath *paths, [namespace-bindings, variable-bindings, custom-handler-class]
-      #
-      # Search this node for XPath +paths+. +paths+ must be one or more XPath
-      # queries.
-      #
-      #   node.xpath('.//title')
-      #
-      # A hash of namespace bindings may be appended. For example:
-      #
-      #   node.xpath('.//foo:name', {'foo' => 'http://example.org/'})
-      #   node.xpath('.//xmlns:name', node.root.namespaces)
-      #
-      # A hash of variable bindings may also be appended to the namespace bindings. For example:
-      #
-      #   node.xpath('.//address[@domestic=$value]', nil, {:value => 'Yes'})
-      #
-      # Custom XPath functions may also be defined.  To define custom
-      # functions create a class and implement the function you want
-      # to define.  The first argument to the method will be the
-      # current matching NodeSet.  Any other arguments are ones that
-      # you pass in.  Note that this class may appear anywhere in the
-      # argument list.  For example:
-      #
-      #   node.xpath('.//title[regex(., "\w+")]', Class.new {
-      #     def regex node_set, regex
-      #       node_set.find_all { |node| node['some_attribute'] =~ /#{regex}/ }
-      #     end
-      #   }.new)
-      #
-      def xpath *paths
-        return NodeSet.new(document) unless document
-
-        paths, handler, ns, binds = extract_params(paths)
-
-        sets = paths.map { |path|
-          ctx = XPathContext.new(self)
-          ctx.register_namespaces(ns)
-          path = path.gsub(/xmlns:/, ' :') unless Nokogiri.uses_libxml?
-
-          binds.each do |key,value|
-            ctx.register_variable key.to_s, value
-          end if binds
-
-          ctx.evaluate(path, handler)
-        }
-        return sets.first if sets.length == 1
-
-        NodeSet.new(document) do |combined|
-          sets.each do |set|
-            set.each do |node|
-              combined << node
-            end
-          end
-        end
-      end
-
-      ###
-      # call-seq: css *rules, [namespace-bindings, custom-pseudo-class]
-      #
-      # Search this node for CSS +rules+. +rules+ must be one or more CSS
-      # selectors. For example:
-      #
-      #   node.css('title')
-      #   node.css('body h1.bold')
-      #   node.css('div + p.green', 'div#one')
-      #
-      # A hash of namespace bindings may be appended. For example:
-      #
-      #   node.css('bike|tire', {'bike' => 'http://schwinn.com/'})
-      #
-      # Custom CSS pseudo classes may also be defined.  To define
-      # custom pseudo classes, create a class and implement the custom
-      # pseudo class you want defined.  The first argument to the
-      # method will be the current matching NodeSet.  Any other
-      # arguments are ones that you pass in.  For example:
-      #
-      #   node.css('title:regex("\w+")', Class.new {
-      #     def regex node_set, regex
-      #       node_set.find_all { |node| node['some_attribute'] =~ /#{regex}/ }
-      #     end
-      #   }.new)
-      #
-      # Note that the CSS query string is case-sensitive with regards
-      # to your document type. That is, if you're looking for "H1" in
-      # an HTML document, you'll never find anything, since HTML tags
-      # will match only lowercase CSS queries. However, "H1" might be
-      # found in an XML document, where tags names are case-sensitive
-      # (e.g., "H1" is distinct from "h1").
-      #
-      def css *rules
-        rules, handler, ns, binds = extract_params(rules)
-
-        prefix = "#{implied_xpath_context}/"
-
-        rules = rules.map { |rule|
-          CSS.xpath_for(rule, :prefix => prefix, :ns => ns)
-        }.flatten.uniq + [ns, handler, binds].compact
-
-        xpath(*rules)
-      end
-
       ###
       # Search this node's immediate children using CSS selector +selector+
       def > selector
@@ -221,44 +114,16 @@ module Nokogiri
         xpath CSS.xpath_for(selector, :prefix => "./", :ns => ns).first
       end
 
-      ###
-      # Search for the first occurrence of +path+.
-      #
-      # Returns nil if nothing is found, otherwise a Node.
-      def at path, ns = document.root ? document.root.namespaces : {}
-        search(path, ns).first
-      end
-      alias :% :at
-
-      ##
-      # Search this node for the first occurrence of XPath +paths+.
-      # Equivalent to <tt>xpath(paths).first</tt>
-      # See Node#xpath for more information.
-      #
-      def at_xpath *paths
-        xpath(*paths).first
-      end
-
-      ##
-      # Search this node for the first occurrence of CSS +rules+.
-      # Equivalent to <tt>css(rules).first</tt>
-      # See Node#css for more information.
-      #
-      def at_css *rules
-        css(*rules).first
-      end
-
       ###
       # Get the attribute value for the attribute +name+
       def [] name
-        return nil unless key?(name.to_s)
         get(name.to_s)
       end
 
       ###
       # Set the attribute value for the attribute +name+ to +value+
       def []= name, value
-        set name.to_s, value
+        set name.to_s, value.to_s
       end
 
       ###
@@ -271,13 +136,30 @@ module Nokogiri
       def add_child node_or_tags
         node_or_tags = coerce(node_or_tags)
         if node_or_tags.is_a?(XML::NodeSet)
-          node_or_tags.each { |n| add_child_node n }
+          node_or_tags.each { |n| add_child_node_and_reparent_attrs n }
         else
-          add_child_node node_or_tags
+          add_child_node_and_reparent_attrs node_or_tags
         end
         node_or_tags
       end
 
+      ###
+      # Add +node_or_tags+ as the first child of this Node.
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the reparented node (if +node_or_tags+ is a Node), or NodeSet (if +node_or_tags+ is a DocumentFragment, NodeSet, or string).
+      #
+      # Also see related method +add_child+.
+      def prepend_child node_or_tags
+        if first = children.first
+          # Mimic the error add_child would raise.
+          raise RuntimeError, "Document already has a root node" if document? && !node_or_tags.processing_instruction?
+          first.__send__(:add_sibling, :previous, node_or_tags)
+        else
+          add_child(node_or_tags)
+        end
+      end
+
       ###
       # Add +node_or_tags+ as a child of this Node.
       # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
@@ -297,22 +179,9 @@ module Nokogiri
       #
       # Also see related method +before+.
       def add_previous_sibling node_or_tags
-        raise ArgumentError.new("A document may not have multiple root nodes.") if parent.is_a?(XML::Document) && !node_or_tags.is_a?(XML::ProcessingInstruction)
+        raise ArgumentError.new("A document may not have multiple root nodes.") if (parent && parent.document?) && !node_or_tags.processing_instruction?
 
-        node_or_tags = coerce(node_or_tags)
-        if node_or_tags.is_a?(XML::NodeSet)
-          if text?
-            pivot = Nokogiri::XML::Node.new 'dummy', document
-            add_previous_sibling_node pivot
-          else
-            pivot = self
-          end
-          node_or_tags.each { |n| pivot.send :add_previous_sibling_node, n }
-          pivot.unlink if text?
-        else
-          add_previous_sibling_node node_or_tags
-        end
-        node_or_tags
+        add_sibling :previous, node_or_tags
       end
 
       ###
@@ -323,22 +192,9 @@ module Nokogiri
       #
       # Also see related method +after+.
       def add_next_sibling node_or_tags
-        raise ArgumentError.new("A document may not have multiple root nodes.") if parent.is_a?(XML::Document)
-        
-        node_or_tags = coerce(node_or_tags)
-        if node_or_tags.is_a?(XML::NodeSet)
-          if text?
-            pivot = Nokogiri::XML::Node.new 'dummy', document
-            add_next_sibling_node pivot
-          else
-            pivot = self
-          end
-          node_or_tags.reverse_each { |n| pivot.send :add_next_sibling_node, n }
-          pivot.unlink if text?
-        else
-          add_next_sibling_node node_or_tags
-        end
-        node_or_tags
+        raise ArgumentError.new("A document may not have multiple root nodes.") if (parent && parent.document?) && !node_or_tags.processing_instruction?
+
+        add_sibling :next, node_or_tags
       end
 
       ####
@@ -388,9 +244,9 @@ module Nokogiri
         node_or_tags = coerce(node_or_tags)
         children.unlink
         if node_or_tags.is_a?(XML::NodeSet)
-          node_or_tags.each { |n| add_child_node n }
+          node_or_tags.each { |n| add_child_node_and_reparent_attrs n }
         else
-          add_child_node node_or_tags
+          add_child_node_and_reparent_attrs node_or_tags
         end
         node_or_tags
       end
@@ -403,17 +259,22 @@ module Nokogiri
       #
       # Also see related method +swap+.
       def replace node_or_tags
+        # We cannot replace a text node directly, otherwise libxml will return
+        # an internal error at parser.c:13031, I don't know exactly why
+        # libxml is trying to find a parent node that is an element or document
+        # so I can't tell if this is bug in libxml or not. issue #775.
+        if text?
+          replacee = Nokogiri::XML::Node.new 'dummy', document
+          add_previous_sibling_node replacee
+          unlink
+          return replacee.replace node_or_tags
+        end
+
         node_or_tags = coerce(node_or_tags)
+
         if node_or_tags.is_a?(XML::NodeSet)
-          if text?
-            replacee = Nokogiri::XML::Node.new 'dummy', document
-            add_previous_sibling_node replacee
-            unlink
-          else
-            replacee = self
-          end
-          node_or_tags.each { |n| replacee.add_previous_sibling n }
-          replacee.unlink
+          node_or_tags.each { |n| add_previous_sibling n }
+          unlink
         else
           replace_node node_or_tags
         end
@@ -490,7 +351,9 @@ module Nokogiri
       ###
       # Remove the attribute named +name+
       def remove_attribute name
-        attributes[name].remove if key? name
+        attr = attributes[name].remove if key? name
+        clear_xpath_context if Nokogiri.jruby?
+        attr
       end
       alias :delete :remove_attribute
 
@@ -513,6 +376,15 @@ module Nokogiri
       # *this* node.  Returns a XML::NodeSet containing the nodes parsed from
       # +string_or_io+.
       def parse string_or_io, options = nil
+        ##
+        # When the current node is unparented and not an element node, use the
+        # document as the parsing context instead. Otherwise, the in-context
+        # parser cannot find an element or a document node.
+        # Document Fragments are also not usable by the in-context parser.
+        if !element? && !document? && (!parent || parent.fragment?)
+          return document.parse(string_or_io, options)
+        end
+
         options ||= (document.html? ? ParseOptions::DEFAULT_HTML : ParseOptions::DEFAULT_XML)
         if Fixnum === options
           options = Nokogiri::XML::ParseOptions.new(options)
@@ -597,6 +469,16 @@ module Nokogiri
         type == HTML_DOCUMENT_NODE
       end
 
+      # Returns true if this is a Document
+      def document?
+        is_a? XML::Document
+      end
+
+      # Returns true if this is a ProcessingInstruction node
+      def processing_instruction?
+        type == PI_NODE
+      end
+
       # Returns true if this is a Text node
       def text?
         type == TEXT_NODE
@@ -761,12 +643,7 @@ module Nokogiri
       # See Node#write_to for a list of +options+.  For formatted output,
       # use Node#to_xhtml instead.
       def to_html options = {}
-        # FIXME: this is a hack around broken libxml versions
-        return dump_html if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
-
-        options[:save_with] |= SaveOptions::DEFAULT_HTML if options[:save_with]
-        options[:save_with] = SaveOptions::DEFAULT_HTML unless options[:save_with]
-        serialize(options)
+        to_format SaveOptions::DEFAULT_HTML, options
       end
 
       ###
@@ -787,12 +664,7 @@ module Nokogiri
       #
       # See Node#write_to for a list of +options+
       def to_xhtml options = {}
-        # FIXME: this is a hack around broken libxml versions
-        return dump_html if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
-
-        options[:save_with] |= SaveOptions::DEFAULT_XHTML if options[:save_with]
-        options[:save_with] = SaveOptions::DEFAULT_XHTML unless options[:save_with]
-        serialize(options)
+        to_format SaveOptions::DEFAULT_XHTML, options
       end
 
       ###
@@ -835,11 +707,7 @@ module Nokogiri
       #
       # See Node#write_to for a list of +options+
       def write_html_to io, options = {}
-        # FIXME: this is a hack around broken libxml versions
-        return (io << dump_html) if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
-
-        options[:save_with] ||= SaveOptions::DEFAULT_HTML
-        write_to io, options
+        write_format_to SaveOptions::DEFAULT_HTML, io, options
       end
 
       ###
@@ -847,11 +715,7 @@ module Nokogiri
       #
       # See Node#write_to for a list of +options+
       def write_xhtml_to io, options = {}
-        # FIXME: this is a hack around broken libxml versions
-        return (io << dump_html) if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
-
-        options[:save_with] ||= SaveOptions::DEFAULT_XHTML
-        write_to io, options
+        write_format_to SaveOptions::DEFAULT_XHTML, io, options
       end
 
       ###
@@ -898,49 +762,77 @@ module Nokogiri
 
       private
 
-      def extract_params params # :nodoc:
-        # Pop off our custom function handler if it exists
-        handler = params.find { |param|
-          ![Hash, String, Symbol].include?(param.class)
-        }
+      def add_sibling next_or_previous, node_or_tags
+        impl = (next_or_previous == :next) ? :add_next_sibling_node : :add_previous_sibling_node
+        iter = (next_or_previous == :next) ? :reverse_each          : :each
 
-        params -= [handler] if handler
-
-        hashes = []
-        while Hash === params.last || params.last.nil?
-          hashes << params.pop
-          break if params.empty?
+        node_or_tags = coerce node_or_tags
+        if node_or_tags.is_a?(XML::NodeSet)
+          if text?
+            pivot = Nokogiri::XML::Node.new 'dummy', document
+            send impl, pivot
+          else
+            pivot = self
+          end
+          node_or_tags.send(iter) { |n| pivot.send impl, n }
+          pivot.unlink if text?
+        else
+          send impl, node_or_tags
         end
-
-        ns, binds = hashes.reverse
-
-        ns ||= document.root ? document.root.namespaces : {}
-
-        [params, handler, ns, binds]
+        node_or_tags
       end
 
-      def coerce data # :nodoc:
-        return data                    if data.is_a?(XML::NodeSet)
-        return data.children           if data.is_a?(XML::DocumentFragment)
-        return fragment(data).children if data.is_a?(String)
+      def to_format save_option, options
+        # FIXME: this is a hack around broken libxml versions
+        return dump_html if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
 
-        if data.is_a?(Document) || data.is_a?(XML::Attr) || !data.is_a?(XML::Node)
-          raise ArgumentError, <<-EOERR
-Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
-(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
-          EOERR
-        end
-
-        data
+        options[:save_with] = save_option unless options[:save_with]
+        serialize(options)
       end
 
-      def implied_xpath_context
-        "./"
+      def write_format_to save_option, io, options
+        # FIXME: this is a hack around broken libxml versions
+        return (io << dump_html) if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
+
+        options[:save_with] ||= save_option
+        write_to io, options
       end
 
       def inspect_attributes
         [:name, :namespace, :attribute_nodes, :children]
       end
+
+      def coerce data # :nodoc:
+        case data
+        when XML::NodeSet
+          return data
+        when XML::DocumentFragment
+          return data.children
+        when String
+          return fragment(data).children
+        when Document, XML::Attr
+          # unacceptable
+        when XML::Node
+          return data
+        end
+
+        raise ArgumentError, <<-EOERR
+Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
+(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
+        EOERR
+      end
+
+      def implied_xpath_contexts # :nodoc:
+        [".//"]
+      end
+
+      def add_child_node_and_reparent_attrs node # :nodoc:
+        add_child_node node
+        node.attribute_nodes.find_all { |a| a.name =~ /:/ }.each do |attr_node|
+          attr_node.remove
+          node[attr_node.name] = attr_node.value
+        end
+      end
     end
   end
 end
