@@ -1,3 +1,5 @@
+# -*- coding: utf-8 -*-
+
 require "helper"
 
 module Nokogiri
@@ -11,6 +13,21 @@ module Nokogiri
         assert_equal 'world', doc.root['abcDef']
       end
 
+      def test_builder_multiple_nodes
+        Nokogiri::XML::Builder.new do |xml|
+          0.upto(10) do
+            xml.text "test"
+          end
+        end
+      end
+
+
+      def test_builder_with_utf8_text
+        text = "test ﺵ "
+        doc = Nokogiri::XML::Builder.new(:encoding => "UTF-8") { |xml| xml.test text }.doc
+        assert_equal text, doc.content
+      end
+
       def test_builder_escape
         xml = Nokogiri::XML::Builder.new { |x|
           x.condition "value < 1", :attr => "value < 1"
@@ -103,6 +120,21 @@ module Nokogiri
         assert_equal 'bar', doc.at('foo|baz', 'foo' => 'bar').namespace.href
       end
 
+      def test_dtd_in_builder_output
+        builder = Nokogiri::XML::Builder.new do |xml|
+          xml.doc.create_internal_subset(
+                                         'html',
+                                         "-//W3C//DTD HTML 4.01 Transitional//EN",
+                                         "http://www.w3.org/TR/html4/loose.dtd"
+                                         )
+          xml.root do
+            xml.foo
+          end
+        end
+        assert_match(/<!DOCTYPE html PUBLIC "-\/\/W3C\/\/DTD HTML 4.01 Transitional\/\/EN" "http:\/\/www.w3.org\/TR\/html4\/loose.dtd">/,
+                     builder.to_xml)
+      end
+
       def test_specify_namespace_nested
         b = Nokogiri::XML::Builder.new { |xml|
           xml.root('xmlns:foo' => 'bar') do
@@ -120,11 +152,27 @@ module Nokogiri
         assert_equal 'bar', doc.at('foo|baz', 'foo' => 'bar').namespace.href
       end
 
+      def test_specified_namespace_postdeclared
+        doc = Nokogiri::XML::Builder.new { |xml|
+          xml.a do
+            xml[:foo].b("xmlns:foo" => "bar")
+          end
+        }.doc
+        a = doc.at('a')
+        assert_equal({}, a.namespaces)
+
+        b = doc.at_xpath('//foo:b', {:foo=>'bar'})
+        assert b
+        assert_equal({"xmlns:foo"=>"bar"}, b.namespaces)
+        assert_equal("b", b.name)
+        assert_equal("bar", b.namespace.href)
+      end
+
       def test_specified_namespace_undeclared
         Nokogiri::XML::Builder.new { |xml|
           xml.root do
             assert_raises(ArgumentError) do
-              xml[:foo]
+              xml[:foo].bar
             end
           end
         }
@@ -209,6 +257,25 @@ module Nokogiri
         assert_equal ["bbb","ccc"], builder.doc.at_css("aaa").children.collect(&:name)
       end
 
+      def test_raw_xml_append_with_namespaces
+        doc = Nokogiri::XML::Builder.new do |xml|
+          xml.root("xmlns:foo" => "x", "xmlns" => "y") do
+            xml << '<Element foo:bar="bazz"/>'
+          end
+        end.doc
+
+        el = doc.at 'Element'
+        assert_not_nil el
+
+        assert_equal 'y', el.namespace.href
+        assert_nil el.namespace.prefix
+
+        attr = el.attributes["bar"]
+        assert_not_nil attr
+        assert_not_nil attr.namespace
+        assert_equal "foo", attr.namespace.prefix
+      end
+
       def test_cdata
         builder = Nokogiri::XML::Builder.new do
           root {
@@ -238,7 +305,33 @@ module Nokogiri
           builder.to_xml.gsub(/\n/, ''))
       end
 
-    private
+      def test_builder_can_inherit_parent_namespace
+        builder = Nokogiri::XML::Builder.new
+        builder.products {
+          builder.parent.default_namespace = "foo"
+          builder.product {
+            builder.parent.default_namespace = nil
+          }
+        }
+        doc = builder.doc
+        ['product', 'products'].each do |n|
+          assert_equal doc.at_xpath("//*[local-name() = '#{n}']").namespace.href, 'foo'
+        end
+      end
+
+      def test_builder_can_handle_namespace_override
+        builder = Nokogiri::XML::Builder.new
+        builder.products('xmlns:foo' => 'bar') {
+          builder.product('xmlns:foo' => 'baz')
+        }
+
+        doc = builder.doc
+        assert_equal doc.at_xpath("//*[local-name() = 'product']").namespaces['xmlns:foo'], 'baz'
+        assert_equal doc.at_xpath("//*[local-name() = 'products']").namespaces['xmlns:foo'], 'bar'
+        assert_nil doc.at_xpath("//*[local-name() = 'products']").namespace
+      end
+
+      private
 
       def namespaces_defined_on(node)
         Hash[*node.namespace_definitions.collect{|n| ["xmlns:" + n.prefix, n.href]}.flatten]
